%{
/*
 * Pfe2mf - lexical analysis
 *
 * Copyright (c) 1992 by Warren Toomey.
 *
 * You may distribute this code for free (or a nominal copying fee) if you
 * do not alter the source code at all. If you want to distribute a changed
 * version, you should distribute a set of patches along with the original.
 *
 * You may not sell this source code or profit in any way from this code.
 */

#include <stdio.h>
#include "ytab.h"
extern YYSTYPE yylval;
extern char *Pfefile;

#ifdef OLDFLEX			/* We need our own yylineno */
int yylineno=1;
#endif

#ifdef DEBUG			/* debugging version - if assert ok */
# include <assert.h>
# define token(x)	(int) "x"

int yynerrs = 0;

main()
 {
  char *p;

  assert(sizeof(int)>=sizeof(char *));

  while (p=(char *)yylex())
	printf("%s is \"%s\"\n",p ,yytext);
 }

# else				/* not DEBUG - production version */
# define token(x)	x
#endif

#define END(v)	(v-1 + sizeof v / sizeof v[0])
static int screen();
%}

letter			[a-zA-Z_]
digit			[0-9]
blank			[ \t]
eoln			[\n]
minus			"-"
dot			"."


%%

{letter}+				return screen();

\/{dot}?{letter}+			{ yylval.name= yytext;
					  return token(IDENTIFIER);
					}

{minus}?{digit}+			{ yylval.value=atoi(yytext);
					  return token(NUMBER);
					}

{eoln}					{
#ifdef OLDFLEX
					yylineno++;
#endif
					}

{blank}+				;

%%

/*
 *	reserved word screener
 */

static struct rwtable {			/* reserved word table */
	char *rw_name;			/* representation */
	int rw_yylex;			/* yylex() value */
    } rwtable[]=			/* sorted */
       {
	"callsubr", token(CALLSUBR),
	"closepath", token(CLOSEPATH),
	"end", token(ENDFILE),
	"endchar", token(ENDCHAR),
	"hlineto", token(HLINETO),
	"hmoveto", token(HMOVETO),
	"hsbw", token(HSBW),
	"hstem", token(HSTEM),
	"hvcurveto", token(HVCURVETO),
	"rlineto", token(RLINETO),
	"rmoveto", token(RMOVETO),
	"rrcurveto", token(RRCURVETO),
	"seac", token(SEAC),
	"vhcurveto", token(VHCURVETO),
	"vlineto", token(VLINETO),
	"vmoveto", token(VMOVETO),
	"vstem", token(VSTEM),
    };

static int screen()
 {
  struct rwtable *low = rwtable,
		 *high = END(rwtable),
		 *mid;
  int c;

/* fprintf(stderr,"Got lex token %s\n",yytext); */

  while (low<=high)
   {
    mid= low + (high-low)/2;
    if ((c = strcmp(mid->rw_name, yytext))==0)
	return mid->rw_yylex;
    else if (c<0)
	low= mid + 1;
    else
	high= mid - 1;
   }
  fprintf(stderr,"Bad lex command >%s<\n",yytext);
 }

int yyerror(err)
 char *err;
 {
  fprintf(stderr,"%s: line %d %s at %s\n",Pfefile, yylineno, err, yytext);
 }
