%{
#include <stdio.h>
#define END(v)  (v-1 + sizeof v / sizeof v[0])
extern int yylineno;

/* Pfe2mf - syntax analysis
 *
 * Copyright (c) 1992 by Warren Toomey.
 *
 * You may distribute this code for free (or a nominal copying fee) if you
 * do not alter the source code at all. If you want to distribute a changed
 * version, you should distribute a set of patches along with the original.
 *
 * You may not sell this source code or profit in any way from this code.
 */

/* Every stroke in a path consists of a node of this type.
 * The id indicates what stroke type it is. The closepath
 * field is TRUE for the last stroke in a path.
 * Id's of type LINE use the start & end fields. Id's of
 * type CURVE use the start, end and bc [Bezier control]
 * fields.
 */
struct pathstruct
	{
	  int startx;
	  int starty;
	  int endx;
	  int endy;
	  int bc1x;
	  int bc1y;
	  int bc2x;
	  int bc2y;
	  int id;
	  int closepath;
	  struct pathstruct *next;
	};
#define LINE	1
#define CURVE	2

/* Global variables.
 */
int X=0,Y=0;			/* Current X,Y position */
int Unknown_chcode=128;		/* Used when we can't find a TeX equivalent */
char CHname[100];		/* Character's name */
int dummyi, dummyj;		/* Dummy variables */
int Char_code;			/* This character's char code */
int Leftgap;			/* Left gap as given from `hsbw' command */
int Width;			/* Character's width */
int Height;			/* Character's height - from afmfile */
int Depth;			/* Character's depth - from afmfile */
int Space;			/* Inter-word space - from afmfile */
int Stretch;			/* Inter-word stretch - from afmfile */
int Shrink;			/* Inter-word shrink - from afmfile */
int Fi=0, Fl=0, Ff=0;		/* True if we've got these characters */
int Ffi=0, Ffl=0;		/* True if we've got these characters */
int Howfill;			/* Hint as how to fill the character */
struct pathstruct *Head,
		  *Tail,
		  *This;	/* Start, end, current node of stroke list */
char *Afmfile=NULL;		/* Name of afmfile */
char *Pfefile=NULL;		/* Name of pfefile */
int Allchars=0;			/* If 1, we parse all chars. Otherwise, just */
				/* the ones with TeX equivalents. */
int Afmparsed=0;		/* 1 is we parsed the afmfile ok. */
char *FontName=NULL;		/* The name of the font */
%}

%union {
	 int value;
	 char *name;
	}

%token <name> IDENTIFIER
%token <value> NUMBER
%token CALLSUBR
%token CLOSEPATH
%token ENDFILE
%token ENDCHAR
%token HLINETO
%token HMOVETO
%token HSBW
%token HSTEM
%token HVCURVETO
%token RLINETO
%token RMOVETO
%token RRCURVETO
%token SEAC
%token VHCURVETO
%token VLINETO
%token VMOVETO
%token VSTEM

%%

Pfbfile
	: Letters ENDFILE
		{ 			/* Finally finished!! */
		  giveligtable();	/* Print any ligtable */
					/* Write out the inter */
					/* word space stuff */
		  printf("\nfont_normal_space %dap#;\n", Space);
		  printf("font_normal_stretch %dap#;\n", Stretch);
		  printf("font_normal_space %dap#;\n", Shrink);
		  printf("\n\n\n%% End of output from pfe2mf\n");
		  exit(0);
		}
	;

Letters
	: Letter
	| Letters Letter
	;

Letter
	: Identifier Hbsw ENDCHAR
	| Identifier Hbsw Commands ENDCHAR
		{ if (Allchars || (Char_code>=0 && Char_code<128))
			givecharacter();
		  else freecharacter();
		 }
	| Identifier Hbsw Seaccommand
	;

Commands
	: Command
	| Commands Command
	;

Command
	: Stemcommand
	| Pathcommand
	| Othercommand
	;

Pathcommand
	: Movecommand Strokecommands CLOSEPATH
		{ This->closepath= 1; }
	| CLOSEPATH
	;

Strokecommands
	: Strokecommand
	| Strokecommands Strokecommand
	;

Strokecommand
	: NUMBER HLINETO
		{
		  newnode();
		  X+= $1;
		  This->endx= X; This->endy= Y;
		  This->id= LINE;
		}
	| NUMBER VLINETO
		{
		  newnode();
		  Y+= $1;
		  This->endx= X; This->endy= Y;
		  This->id= LINE;
		}
	| NUMBER NUMBER RLINETO
		{
		  newnode();
		  X+= $1; Y+= $2;
		  This->endx= X; This->endy= Y;
		  This->id= LINE;
		}
	| NUMBER NUMBER NUMBER NUMBER HVCURVETO
		{
		  newnode();
		  X+= $1;
		  This->bc1x= X; This->bc1y= Y;
		  X+= $2; Y+= $3;
		  This->bc2x= X; This->bc2y= Y;
		  Y+= $4;
		  This->endx= X; This->endy= Y;
		  This->id= CURVE;
		}
	| NUMBER NUMBER NUMBER NUMBER VHCURVETO
		{
		  newnode();
		  Y+= $1;
		  This->bc1x= X; This->bc1y= Y;
		  X+= $2; Y+= $3;
		  This->bc2x= X; This->bc2y= Y;
		  X+= $4;
		  This->endx= X; This->endy= Y;
		  This->id= CURVE;
		}
	| NUMBER NUMBER NUMBER NUMBER NUMBER NUMBER RRCURVETO
		{
		  newnode();
		  X+= $1; Y+= $2;
		  This->bc1x= X; This->bc1y= Y;
		  X+= $3; Y+= $4;
		  This->bc2x= X; This->bc2y= Y;
		  X+= $5; Y+= $6;
		  This->endx= X; This->endy= Y;
		  This->id= CURVE;
		}

	;

Movecommand
	: NUMBER HMOVETO
		{ X+= $1; }
	| NUMBER VMOVETO
		{ Y+= $1; }
	| NUMBER NUMBER RMOVETO
		{ X+= $1; Y+= $2; }
	;

Stemcommand
	: NUMBER NUMBER VSTEM
	| NUMBER NUMBER HSTEM
	;

Othercommand
	: NUMBER CALLSUBR
	| NUMBER NUMBER CALLSUBR
	;

Hbsw
	: NUMBER NUMBER HSBW
		{
		  Leftgap= $1; Width= $2;
		  X=$1; Y=0;
		}
	;

Seaccommand
	: NUMBER NUMBER NUMBER NUMBER NUMBER SEAC
	;

Identifier
	: IDENTIFIER
		{ chscreen($1);
		  Head= Tail= NULL;
		}
	;
%%
/* Get a line of input into the buffer.
 * Do not exceed the buffer size. Return
 * the number of chars in the buffer.
 * Ignore leading whitespace.
 */
int getline(buf,size)
 char *buf;
 int size;
 {
  int ch;
  int numgot=0;

  size--;
  while ((ch=getchar())!=EOF && (ch==' ' || ch=='\t' || ch=='\n'))
   if (ch=='\n') yylineno++;

  while(ch!=EOF && ch!='\n' && numgot<size)
   {
    buf[numgot++]= ch; ch=getchar();
   }
  buf[numgot++]='\0'; yylineno++;
  return(numgot);
 }

void usage()
 { fprintf(stderr,"Usage: pfe2mf [-a] [afmfile] pfefile\n"); exit(1); 
 }


/* Wade thru the first half of the pfe file
 *
 */
parsepfe()
 {
  char inbuf[500];			/* Should be enough! */
  int i,amt;
  int percentleft=1;

  printf("%% Information from the pfe file:\n%%\n");

  while(1)
   {
    amt= getline(inbuf,500);
    if (amt==0) exit(1);
    if (percentleft && inbuf[0]=='%') puts(inbuf);
    else percentleft=0;
    if (strstr(inbuf,"CharString")) return;	/* End of top half */
    if (!strncmp(inbuf,"/FontName",9) && FontName==NULL)
     {
      FontName= (char *)malloc(strlen(inbuf)-9);
      if (FontName)
	for (i=0,amt=11; inbuf[amt]!=' ' && inbuf[amt]!='\0';)
		FontName[i++]= inbuf[amt++];
	FontName[i]='\0';
      continue;
     }
    if (!strncmp(inbuf,"/version",8))
	{ inbuf[0]='%'; puts(inbuf); continue; }
    if (!strncmp(inbuf,"/Notice",7))
	{ inbuf[0]='%'; puts(inbuf); continue; }
    if (!strncmp(inbuf,"/FullName",9))
	{ inbuf[0]='%'; puts(inbuf); continue; }
    if (!strncmp(inbuf,"/FamilyName",11))
	{ inbuf[0]='%'; puts(inbuf); continue; }
    if (!strncmp(inbuf,"/Weight",7))
	{ inbuf[0]='%'; puts(inbuf); continue; }
   }
 }

/* Print out the header of the Metafont Code.
 */
void printheader()
 {
  
  printf("%%\n%%\n%% The following constants convert Adobe Type 1 binary\n");
  printf("%% point size into Metafont values. You may\n");
  printf("%% want to alter these values to get the\n");
  printf("%% correct point size output by Metafont\n%%\n%%\n");

  printf("ap#:= pt#/95;\nap= pt/95;\n\n");

  if (FontName) printf("font_identifier \"%s\";\n",FontName);
  printf("font_size 10pt#;\n\n\n");
 }


static struct chrtable {	/* character lookup table */
	char *ch_name;		/* character's Adobe name */
	int ch_code;		/* equivalent TeX code */
	int how_fill;		/* plus how to fill hint */
    } chtable[]=		/* sorted */
       {
	".notdef", -1, 1,
	"A", 65, 0,
	"AE", 29, 0,
	"B", 66, 0,
	"C", 67, 1,
	"D", 68, 0,
	"E", 69, 1,
	"F", 70, 1,
	"G", 71, 1,
	"H", 72, 1,
	"I", 73, 1,
	"J", 74, 1,
	"K", 75, 1,
	"L", 76, 1,
	"M", 77, 1,
	"N", 78, 1,
	"O", 79, 0,
	"OE", 30, 0,
	"Oslash", 31, 0,
	"P", 80, 0,
	"Q", 81, 0,
	"R", 82, 0,
	"S", 83, 1,
	"T", 84, 1,
	"U", 85, 1,
	"V", 86, 1,
	"W", 87, 1,
	"X", 88, 1,
	"Y", 89, 1,
	"Z", 90, 1,
	"a", 97, 0,
	"acute", 19, 1,
	"ae", 26, 0,
	"ampersand", 38, 0,
	"asterisk", 42, 1,
	"at", 64, 0,
	"b", 98, 0,
	"bracketleft", 91, 1,
	"bracketright", 93, 1,
	"breve", 21, 1,
	"c", 99, 1,
	"caron", 20, 1,
	"cedilla", 24, 1,
	"circumflex", 94, 1,
	"colon", 58, 1,
	"comma", 44, 1,
	"d", 100, 0,
	"dieresis", 127, 1,
	"dollar", 36, 0,
	"dotaccent", 95, 1,
	"e", 101, 0,
	"eight", 56, 0,
	"emdash", 124, 1,
	"endash", 123, 1,
	"equal", 61, 1,
	"exclam", 33, 1,
	"f", 102, 1,
	"ff", 11, 1,
	"ffi", 14, 1,
	"ffl", 15, 1,
	"fi", 12, 1,
	"five", 53, 1,
	"fl", 13, 1,
	"four", 52, 0,
	"g", 103, 0,
	"germandbls", 25, 1,
	"grave", 18, 1,
	"greater", 62, 1,
	"h", 104, 1,
	"hyphen", 45, 1,
	"i", 105, 1,
	"j", 106, 1,
	"k", 107, 1,
	"l", 108, 1,
	"less", 60, 1,
	"m", 109, 1,
	"macron", 22, 1,
	"n", 110, 1,
	"nine", 57, 0,
	"numbersign", 35, 0,
	"o", 111, 0,
	"oe", 27, 0,
	"one", 49, 1,
	"oslash", 28, 0,
	"p", 112, 0,
	"parenleft", 40, 1,
	"parenright", 41, 1,
	"percent", 37, 0,
	"period", 46, 1,
	"plus", 43, 1,
	"q", 113, 0,
	"question", 63, 1,
	"quotedbl", 125, 1,
	"quotedblleft", 92, 1,
	"quotedblright", 34, 1,
	"quoteleft", 96, 1,
	"quoteright", 39, 1,
	"r", 114, 1,
	"ring", 23, 0,
	"s", 115, 1,
	"semicolon", 59, 1,
	"seven", 55, 1,
	"six", 54, 0,
	"slash", 47, 1,
	"t", 116, 1,
	"three", 51, 1,
	"tilde", 126, 1,
	"two", 50, 1,
	"u", 117, 1,
	"v", 118, 1,
	"w", 119, 1,
	"x", 120, 1,
	"y", 121, 1,
	"z", 122, 1,
	"zero", 48, 0,
    };

int chscreen(name)	/* Get charcode from Identifier */
 char *name;
 {
  struct chrtable *low = chtable,
		 *high = END(chtable),
		 *mid;
  int c;

  name++;			/* Skip leading slash */
  strcpy(CHname, name);

  while (low<=high)
   {
    mid= low + (high-low)/2;
    if ((c = strcmp(mid->ch_name, name))==0)
	{ Char_code= mid->ch_code;
	  Howfill= mid->how_fill;
	  return;
	}
    else if (c<0)
	low= mid + 1;
    else
	high= mid - 1;
   }
#ifdef DEBUG
  fprintf(stderr,"Unknown charname %s\n", name);
#endif
  Char_code= Unknown_chcode++;
  if (Unknown_chcode==256) Unknown_chcode--;
  Howfill=0;
  return;
 }

/* Create a new node pointed at by This. Add This to the
 * tail of the Pathlist
 */
void newnode()
 {
  This= (struct pathstruct *)malloc(sizeof(struct pathstruct));
  if (This==NULL) { fprintf(stderr,"Malloc failed in  newnode\n"); exit(1); }

  if (Head==NULL) Head=Tail=This;
  else { Tail->next= This; Tail= This; }
  This->startx= X;
  This->starty= Y;
  This->closepath= 0;
  This->next= NULL;
 }

/* Deduce the height and depth of a character based
 * on the maximum y co-ordinates in the paths.
 * We add on 3% to each for good measure.
 *
 * While we're here, we calculate the interword space, stretch
 * and shrink, using the widths for x, M, and i respectively.
 */
void gethtdepth(chcode)
 int chcode;
 {
  Height=0; Depth=0;

  for (This= Head; This; This=This->next)
   {
    if (This->starty > Height) Height= This->starty;
    if (This->endy > Height) Height= This->endy;
    if (This->starty < Depth) Depth= This->starty;
    if (This->endy < Depth) Depth= This->endy;
   }

  Height+= Height * 3 / 100;
  Depth= -Depth; Depth+= Depth * 3 / 100;
 
  switch(chcode)
   {
	case 120: Space= Width;			/* x */
		  break;
	case 77: Stretch= Width;		/* M */
		  break;
	case 105: Shrink= Width;		/* i */
   }
 }


/* Free a character - don't print out the code
 */
int freecharacter()
 {
  for (This= Head; This;)
   {
    Tail= This; This=This->next; free(Tail);
   }
 }


/* Print out the Metafont code for the
 * character.
 */
int givecharacter()
 {
  int startnode=1;

  printf("beginchar(%d, ",Char_code);
  if (!Afmparsed)			/* Have we got Height & Depth? */
    gethtdepth(Char_code);

  switch(Char_code)			/* Check for the ff ligatures */
   {
	case 11:  Ff= 1;
		  break;
	case 12:  Fi= 1;
		  break;
	case 13:  Fl= 1;
		  break;
	case 14:  Ffi= 1;
		  break;
	case 15:  Ffl= 1;
   }

  printf(" %dap#, %dap#, %dap#); ",Width, Height, Depth);
  printf("\"The letter %s\";\n", CHname);

  for (This= Head; This;)
   {
/* The following is a heuristic to determine if we
 * should fill or unfill. We fill for the first
 * stroke and unfill the others, unless Howfill is 1.
 * This should be replaced with an algorithm for
 * determining if a path is clockwise or not.
 */

    if (startnode)			/* We must issue a fill command */
      { startnode=0;
	if (This == Head || Howfill) printf("fill\n");
	else printf("unfill\n");
      }

    switch(This->id)
     {
      case LINE:
		printf(" (%dap,%dap) -- ",This->startx,This->starty);
		printf("(%dap,%dap) &\n",This->endx,This->endy);
		break;
      case CURVE:
		printf(" (%dap,%dap) .. ",This->startx,This->starty);
		printf("controls (%dap,%dap) ",This->bc1x,This->bc1y);
		printf("and (%dap,%dap) ",This->bc2x,This->bc2y);
		printf(".. (%dap,%dap) &\n",This->endx,This->endy);
		break;
      default:
		fprintf(stderr,"Internal error - bad pathnode id %d\n",
			This->id);
		exit(1);
     }
    if (This->closepath) { printf("cycle;\n"); startnode=1; }

    Tail= This; This=This->next; free(Tail);
   }
  printf("endchar;\n");
 }

/* Print out any ligature table if we have any ff chars
 */
int giveligtable()
 {
  int i,j;

  i= Fi + Ff + Fl;
  j= Ffi + Ffl;
  if (i)				/* We can make a ligtable */
   {
    printf("\nligtable \"f\": ");

    if (Fi)
      { printf("\"i\"=:12");
        if (--i) printf(", "); else printf(";\n");
      }
    if (Ff)
      { printf("\"f\"=:11");
    	if (--i) printf(", "); else printf(";\n");
      }
    if (Fl)
      { printf("\"l\"=:13");
    	if (--i) printf(", "); else printf(";\n");
      }

    if (j)
     {
      printf("ligtable 11: ");
      if (Ffi)
	{ printf("\"i\"=:14");
      	  if (--j) printf(", "); else printf(";\n");
        }
      if (Ffl)
	{ printf("\"l\"=:15");
	  if (--j) printf(", "); else printf(";\n");
        }
     }
   }
 }

int main(argc, argv)
 int argc;
 char *argv[];
 {
  int i;

  switch(argc)
   {
    case 2:	Pfefile= argv[1];
		break;
    case 3:	if (!strcmp("-a",argv[1])) Allchars= 1;
		else Afmfile= argv[1];
		Pfefile= argv[2];
		break;
    case 4:	if (!strcmp("-a",argv[1])) Allchars= 1;
		else usage();
		Afmfile= argv[2];
		Pfefile= argv[3];
		break;
    default:	usage();
   }

  if (freopen(Pfefile, "r", stdin)==NULL)
    { fprintf(stderr, "Couldn't open %s\n", Pfefile); exit(1); }
   
#ifdef OLDFLEX
/* Other lexers may need this too :-( */

  setbuf(stdin, NULL);
#endif
  printf("%%%% Metafont code produced by pfe2mf\n%%%% from file");
  if (Afmfile) printf("s %s and",Afmfile);
  printf(" %s.\n%%\n",Pfefile);

  parsepfe();
  printheader();
  i= yyparse();
#ifdef DEBUG
  fprintf(stderr,"yyparse() == %d\n",i);
#endif
 }
