#include <stdio.h>
#include <ctype.h>

/*
 *  Convert a PFB file to readable ASCII.
 *
 *  A .pfb file is a sequence of packets.  Each packet starts with byte
 *  0x80.  The second byte in the packet gives the type of packet: 1
 *  means it's a packet of ascii data which should be sent out as is
 *  (except for translating \r to the appropriate end-of-line
 *  character(s)); 2 means it's a packet of binary data which should be
 *  translated into hex; 3 means EOF.  For types 1 and 2, the type byte
 *  is followed by four bytes giving the length of the packet, least
 *  significant first.
 *
 *  It get's worse after that. If I ever meet the guy who designed
 *  the Type-1 format, he's gonna be sorry.
 *
 * Copyright (c) 1992 by Warren Toomey.		wkt@csadfa.cs.adfa.oz.au
 *
 * You may distribute this code for free (or a nominal copying fee) if you
 * do not alter the source code at all. If you want to distribute a changed
 * version, you should distribute a set of patches along with the original.
 *
 * You may not sell this source code or profit in any way from this code.
 */
#ifdef MSDOS
# define RB "rb"
#else
# define RB "r"
#endif

static unsigned short buffer = 0xd971;


/* Get character from the input file. We decode both the Type-1
 * binary format and the eexec format.
 */
int
Getchar()
{
  static int ignore = 4;
  static int getstate = 0;
  static unsigned int count;
  int ch, out;


  switch (getstate)
  {
    case 0:			/* Don't know if binary or ASCII yet */
      if ((ch = getchar()) == EOF)
	return (EOF);
      if (ch == 0x80)
      {
	getstate = getchar();
	count = getchar();
	count += (getchar() << 8);
	count += (getchar() << 16);
	count += (getchar() << 24);
	if (getstate == 2)
	  goto c2;		/* Yuk, a goto */
	else
	  goto c1;		/* Yuk, a goto */
      }
      else
      { if (ch == '\r')
	  ch = '\n';
	return (ch);
      }

    case 1:
  c1:				/* ASCII */
      ch = getchar();
      if (ch == EOF)
	return (EOF);
      if (ch == '\r')
	ch = '\n';
      if (--count == 0)
	getstate = 0;
      return (ch);
    case 2:
  c2:				/* Binary */
      ch = getchar();
      if (ch == EOF)
	return (EOF);
      out = ch ^ (buffer >> 8);
      buffer = (unsigned short)((long)(ch + buffer) * 0xce6dL + 0x58bfL);
      if (--count == 0)
	getstate = 0;
      if (ignore > 0)
      {
	ignore--;
	goto c2;
      }				/* Yuk, a goto */
      else
	return (out);
  }
}

#define STKSIZE 40

main(argc, argv)
  int argc;
  char *argv[];
{
  int ch;
  int ND = 0;
  int num = 0;
  int state = 0;
  int z = 0;
  char stk[STKSIZE];

  switch (argc)
  {
    case 1:
      break;
    case 2:
      if (freopen(argv[1], RB, stdin) == NULL)
      {
	fprintf(stderr, "Could not open %s\n", argv[1]);
	exit(1);
      }
      break;
    default:
      fprintf(stderr, "Usage: pfb2pfe [pfbfile]\n");
      exit(1);
  }

  while ((ch = Getchar()) != EOF)
  {
   if (ch=='\r') ch='\n';

/*
 * I knew the 3rd year compiler course I did would come in handy one day!!
 *
 * We grep for number RD, number chars after that is 
 * handled by parsecharstring.
 */

    switch (state)
    {
      case 0:
	if (isdigit(ch))
	{
      firstdigit:
	  num = ch - '0';
	  goto pushchar;	/* Yuk, a goto */
	}
	putchar(ch);
	break;
      case 1:
	if (ch == ' ')
	  goto pushchar;	/* Yuk, a goto */
	if (isdigit(ch))
	{
	  num = 10 * num + ch - '0';
	  state--;
	  goto pushchar;	/* Yuk, a goto */
	}
	goto killstack;		/* Yuk, a goto */
      case 2:
	if (ch != 'R')
	  goto killstack;	/* Yuk, a goto */
    pushchar:
	state++;
	stk[z++] = ch;
	if (z==STKSIZE)		/* Heuristic - we've hit the 00000 at the */
	 exit(0);		/* bottom of the file */
	break;
      case 3:
	if (ch == 'D')
	{
	  state++;
	  break;
	}
    killstack:
	for (num = 0; num < z; num++)
	  putchar(stk[num]);
	z = state = num = 0;
	if (isdigit(ch))
	  goto firstdigit;	/* Yuk, a goto */
	else
	  putchar(ch);
	break;

/* State 4 is where we decode the encrypted
 * commands and numbers
 */

      case 4:
	parsecharstring(num);
	Getchar();
	Getchar();		/* Ignore the 3-char trailer */
	Getchar();
	z = state = 0;
    }
  }
}

static char *cmdlist[32] = {
  "NULL",
  "hstem",
  "NULL",
  "vstem",
  "vmoveto",
  "rlineto",
  "hlineto",
  "vlineto",
  "rrcurveto",
  "closepath",
  "callsubr",
  "return",
  "ESCAPE",
  "hsbw",
  "endchar",
  "NULL",
  "NULL",
  "NULL",
  "NULL",
  "NULL",
  "NULL",
  "rmoveto",
  "hmoveto",
  "NULL",
  "NULL",
  "NULL",
  "NULL",
  "NULL",
  "NULL",
  "NULL",
  "vhcurveto",
  "hvcurveto"
};

unsigned short int getgetbuffer;

int 
Getgetchar()
{
  int c;
  int out;

  c = Getchar();
  out = c ^ (getgetbuffer >> 8);
  getgetbuffer = (unsigned short)((long)(c + getgetbuffer) * 0xce6dL + 0x58bfL);
  return out;
}

int 
parsecharstring(n)
{
  int c;
  int w;

  getgetbuffer = 4330;
  Getgetchar();
  Getgetchar();			/* Ignore 4 random byte header */
  Getgetchar();
  Getgetchar();
  n -= 4;

/* Now determine what the input bytes mean
 */
  while (n > 0)
  {
    c = Getgetchar();
    if (c > 31)			/* It's an integer */
    {
      if (c == 255)
      {
	c = (Getgetchar() << 24);
	c += (Getgetchar() << 16);
	c += (Getgetchar() << 8);
	c += Getgetchar();
	printf("%d ", c);
	n -= 5;
      }
      else if (c > 250)
      {
	w = Getgetchar();
	c = 256 * (c - 251) + w + 108;
	c = -c;
	printf("%d ", c);
	n -= 2;
      }
      else if (c > 246)
      {
	w = Getgetchar();
	c = 256 * (c - 247) + w + 108;
	printf("%d ", c);
	n -= 2;
      }
      else
      {
	printf("%d ", c - 139);
	n--;
      }
    }
    else
    {				/* It's a command */
      if (c != 12)
      {
	n--;
	printf("%s\n", cmdlist[c]);
      }
      else
      {
	n -= 2;
	switch (Getgetchar())
	{
	  case 0:
	    printf("dotsection\n");
	    break;
	  case 1:
	    printf("vstem3\n");
	    break;
	  case 2:
	    printf("hstem3\n");
	    break;
	  case 6:
	    printf("seac\n");
	    break;
	  case 7:
	    printf("sbw\n");
	    break;
	  case 12:
	    printf("div\n");
	    break;
	  case 16:
	    printf("callothersub\n");
	    break;
	  case 17:
	    printf("pop\n");
	    break;
	  case 33:
	    printf("setcurrentpoint\n");
	    break;
	}
      }
    }
  }
}
